#include "common.h"
/* $Header: op_aux.c,v 2.1 89/09/08 11:53:05 gary Locked $ */
#include "instr.h"
#include "op.h"
#include "manager.h"
#include "micro.h"
#include "emitter.h"
/* Because these are ISETL operations, the arguments are on the ISETL
 * stack, not passed as C arguments.  Comments refer to the arguments
 * on the stack.  They are correctly accessed by the ENTER macros.
 * See op.h for an explanation of ENTER, LEAVE, R(X_Op).
 * See instr.h for further description of ISETL operations.
 */


/* (ref,val) ref := val */
void    opAssign()
{
    NO_BIN_ENTER;
    if (trace_debug) {
	mcPrint(R(Left_Op), stdout, verbose);
	show(" := ", stdout);
	mcPrint(R(Right_Op), stdout, verbose);
	mcFlush(stdout);
    }
    mcAssign(R(Left_Op), R(Right_Op));
    BIN_NO_RESULT;
}


/* (x):x,x  Duplicates the top of stack */
void    opDup()
{
    Reg     (ans);

    R(ans) = R(Top - 1);
}


/* () Fetches var index and extends Env */
void    opEnv_Extend()
{
    C_Ref   var;

    var = Get_Ref;

    Environment = mcCons(Nil, Environment);
    Info(Environment) = New_Block(Assoc, 0);
    Var_Index(Info(Environment)) = var;
    TRACE_VARS("Extend by", var);
}


/* ()  Removes an assoc from Env */
void    opEnv_Pop()
{
    Environment = Next(Environment);
}



/* (ref):*   De-reference top of stack */
void    opLoad()
{
    UNARY_ENTER;
    C_Ref   id;

    assert(Kind(R(Only_Op)) == Ref);
    id = Var_Index(Var_Assoc(R(Only_Op)));
    TRACE_VARS("Loading", id);
    if (trace_debug) {
	mcprintf(" = ",0,0);
	mcPrint(Var(R(Only_Op)), stdout, verbose);
	mcFlush(stdout);
    }
    R(ans) = Var(R(Only_Op));
    switch (Kind(R(ans))) {
      case Tuple:
      case Set:
      case Closure:
      case String:
      case Predef:
	if (Watched(id)) {
	    Id_Watched(R(ans)) = id;
	} else {
	    Id_Watched(R(ans)) = 0;
	}
    }
    UNARY_LEAVE;
}


/* ():Atom Push a new atom */
void    opPAtom()
{
    Reg     (new);

    R(new) = New_Block(Atom, 0);
}



/* ():ref Fetches var indx and pushes ref */
void    opPRef()
{
    IPtr     e;
    C_Ref   i;
    Reg     (new);

    R(new) = New_Block(Ref, 0);
    i = Get_Ref;
    TRACE_VARS("Push", i);
    /* Find e where Info(e) = <i,i-value> */
    e = Environment;
    while (Next(e) != Nil && Var_Index(Info(e)) != i) {
	e = Next(e);
    }
    if (Var_Index(Info(e)) != i) {
	Next(e) = New_Block(List, 0);
	e = Next(e);
	Info(e) = New_Block(Assoc, 0);
	Var_Index(Info(e)) = i;
	TRACE_VARS("New Global", i);
    }
    Var_Assoc(R(new)) = Info(e);
}


/* (list): list	 Reverses list */
void    opRev()
{
    UNARY_ENTER;
    R(ans) = mcRev(R(Only_Op));
    UNARY_LEAVE;
}


/* (x,y):y,x  Swaps top two stack elements */
void    opSwap()
{
    IPtr     t;

    t = R(Top);
    R(Top) = R(Top - 1);
    R(Top - 1) = t;
}



/* (low,high):List	low..high	 */
void    opCnt2()
{
    BIN_ENTER;

    switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
      case Pair(Integer, Integer):
	R(ans) = mcCnt(Int_Value(R(Left_Op)),
		       Int_Value(R(Right_Op)),
		       (C_Int) 1);
	break;
      case Pair(Integer, Bignum):
      case Pair(Bignum, Integer):
      case Pair(Bignum, Bignum):
	R(Left_Op) = mpForce(R(Left_Op));
	R(Right_Op) = mpForce(R(Right_Op));
	R(ans) = mpCnt(R(Left_Op), R(Right_Op), BigOne);
	break;

      default:
	RT_ERROR("Bad args in low..high");
    }

    BIN_LEAVE;
}



/* (1st,2nd,3rd): List	1st,2nd..3rd	 */
void    opCnt3()
{
    TRI_ENTER;

    if (Kind(R(First_Op)) == Integer
	&& Kind(R(Second_Op)) == Integer
	&& Kind(R(Third_Op)) == Integer) {

	R(ans) = mcCnt(Int_Value(R(First_Op)),
		       Int_Value(R(Third_Op)),
		       Int_Value(R(Second_Op))
		       - Int_Value(R(First_Op)));
    } else if (
	    (Kind(R(First_Op)) == Integer || Kind(R(First_Op)) == Bignum)
	&& (Kind(R(Second_Op)) == Integer || Kind(R(Second_Op)) == Bignum)
	 && (Kind(R(Third_Op)) == Integer || Kind(R(Third_Op)) == Bignum)
	) {
	Reg     (diff);

	R(First_Op) = mpForce(R(First_Op));
	R(Second_Op) = mpForce(R(Second_Op));
	R(Third_Op) = mpForce(R(Third_Op));
	R(diff) = mpSub(R(Second_Op), R(First_Op));
	R(diff) = mpForce(R(diff));
	R(ans) = mpCnt(R(First_Op), R(Third_Op), R(diff));
    } else {
	RT_ERROR("Bad args in low,next..high");
    }

    TRI_LEAVE;
}



/* (x,L):List cons(x,L) */
void    opCons()
{
    BIN_ENTER;
    assert(Kind(R(Right_Op)) == List);

    R(ans) = mcCons(R(Left_Op), R(Right_Op));
    BIN_LEAVE;
}



/* ()	No operation */
void    opNoOp()
{
}


/* ():List Pushes Nil */
void    opPNil()
{
    Reg     (new);

    R(new) = Nil;
}



/* ():Boolean Pushes True */
void    opPTrue()
{
    Reg     (new);

    R(new) = ITRUE;
}

/* ():Boolean Pushes False */
void    opPFalse()
{
    Reg     (new);

    R(new) = IFALSE;
}

/* ():Undefined Pushes OM */
void    opPOM()
{
    Reg     (new);

    R(new) = OM;
}



/* (x)	Ignores x.  Deletes it from Stack */
void    opPop()
{
    mcPop();
}

